import ibis
import pytest
from packaging.version import parse as vparse

from ibis_substrait.compiler.core import SubstraitCompiler

pa = pytest.importorskip("pyarrow")
pc = pytest.importorskip("pyarrow.compute")
pa_substrait = pytest.importorskip("pyarrow.substrait")

try:
    from ibis.udf.vectorized import elementwise
except ImportError:
    from ibis.legacy.udf.vectorized import elementwise


arrow12 = pytest.mark.skipif(
    vparse(pa.__version__) <= vparse("11.0.0"),
    reason="UDF support added in Arrow 12",
)


def get_table_provider(tbl):
    """Create a table_provider that always returns tbl"""

    def table_provider(names, schema):
        return tbl

    return table_provider


def to_ibis_table(arrow_table, table_name="t1"):
    """Create ibis Table from pyarrow Table"""
    # TODO: use ibis.backends.pyarrow.datatypes.from_pyarrow_schema once we drop ibis 3.x
    return ibis.table(
        zip(arrow_table.schema.names, [str(type) for type in arrow_table.schema.types]),
        name=table_name,
    )


def run_query(plan, tbl):
    query_bytes = plan.SerializeToString()
    result = pa_substrait.run_query(
        # PyArrow wants its bytes in a very specific byte-string
        pa.py_buffer(query_bytes),
        table_provider=get_table_provider(tbl),
    )

    results = result.read_all()
    assert isinstance(results, pa.lib.Table)

    return results


@pytest.fixture
def arrow_table():
    return pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
            "b": [3.4, 3.7, 2.0],
            "c": ["x", "y", "z"],
            "d": [True, False, True],
        }
    )


@arrow12
def test_pyarrow_produces_correct_result(compiler, arrow_table):
    ibis_table = to_ibis_table(arrow_table)
    query = ibis_table  # identity
    plan = compiler.compile(query)
    result = run_query(plan, arrow_table)

    assert result == arrow_table


@arrow12
def test_pyarrow_can_consume_basic_operations(compiler, arrow_table):
    t = to_ibis_table(arrow_table)

    # identity
    query = t
    plan = compiler.compile(query)
    result = run_query(plan, arrow_table)
    assert result == arrow_table

    # mutate, re-use column
    query = t.mutate(a=t.a * 2)
    plan = compiler.compile(query)
    result = run_query(plan, arrow_table)

    assert set(result.column_names) == set(arrow_table.column_names)

    # mutate, add new derived column
    query = t.mutate(e=t.b + 1)
    plan = compiler.compile(query)
    result = run_query(plan, arrow_table)

    assert set(result.column_names) == {*arrow_table.column_names, "e"}


@arrow12
def test_extension_udf():
    def register_pyarrow_udf(udf, registry=None):
        """
        Boilerplate that is needed for the udfs to execute in PyArrow. Nothing to do with ibis.
        """
        import inspect

        try:
            # Ibis 6.x
            from ibis.formats.pyarrow import PyArrowType

            to_pyarrow_type = PyArrowType.from_ibis
        except ImportError:
            # Ibis 4.x, 5.x
            from ibis.backends.pyarrow.datatypes import to_pyarrow_type

        if registry is None:
            registry = pc.function_registry()

        def wrapper(ctx, *args):
            out = udf.func(*args, ctx=ctx)
            return out

        in_types = dict(
            zip(
                inspect.getfullargspec(udf.func).args,
                map(to_pyarrow_type, udf.input_type),
            )
        )
        out_type = to_pyarrow_type(udf.output_type)

        if udf.func.__name__ not in set(registry.list_functions()):
            pc.register_scalar_function(
                wrapper,
                udf.func.__name__,
                {
                    "summary": f"UDF {udf.func.__name__} defined in Ibis",
                    "description": "",
                },
                in_types,
                out_type,
            )

    # type-specifications are for ibis' type-checking
    # part of ibis' lazy evaluation
    @elementwise(input_type=["int64"], output_type="int64")
    def add1(col: pa.Int64Scalar, ctx=None) -> pa.Int64Scalar:
        return pc.call_function("add", [col, 1], memory_pool=ctx.memory_pool)

    # @elementwise defines a UDF that operates element-wise
    @elementwise(input_type=["int64"], output_type="int64")
    def sub1(col: pa.Int64Scalar, ctx=None) -> pa.Int64Scalar:
        # PyArrow.compute code
        return pc.call_function("subtract", [col, 1], memory_pool=ctx.memory_pool)

    # registers the UDF on the execution side with PyArrow.
    registry = pc.function_registry()
    register_pyarrow_udf(add1, registry)
    register_pyarrow_udf(sub1, registry)

    # setting up the ibis table
    t = ibis.table([("a", "int")], name="t")
    # the ibis-substrait compiler requires an ibis' table expression
    query = t.mutate(b=add1(t.a), c=sub1(t.a))

    compiler = SubstraitCompiler(
        # UDF uri for the sake of PyArrow
        # so that it knows where to look to find the UDF
        udf_uri="urn:arrow:substrait_simple_extension_function"
    )

    plan = compiler.compile(query)

    arrow_table = pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
        }
    )

    result = run_query(plan, arrow_table)

    assert result[1].to_pylist() == [2, 3, 4]
    assert result[2].to_pylist() == [0, 1, 2]


@arrow12
def test_acero_unsafe_cast(compiler):
    """
    Regression test to catch any behavior change in how Acero handles casting.

    As of arrow 196fbe5f5e0669b1f8bf71005c8a4c6a4e0fb2da, Acero performs an
    unsafe cast where it previously performed a safe cast. The previous behavior
    would cause the following test to throw pyarrow.ArrowInvalid. See
    https://github.com/apache/arrow/issues/34644 for more detail.
    """
    arrow_table = pa.Table.from_pydict(
        {
            "a": [1.8, 2.2, 3.5],
        }
    )
    t = to_ibis_table(arrow_table)
    query = t.mutate(a=t.a.cast("int"))
    plan = compiler.compile(query)
    result = run_query(plan, arrow_table)

    assert result == pa.Table.from_pydict(
        {
            "a": [1, 2, 3],
        }
    )



import operator

import ibis
import ibis.expr.operations as ops
import pytest

DEFAULT_PREFIX = "https://github.com/substrait-io/substrait/blob/main/extensions"


@pytest.mark.parametrize(
    "two, three",
    [
        (2, 3),
        (ibis.literal(2, "int32"), ibis.literal(3, "int32")),
        (ibis.literal(2, "int32"), ibis.literal(3, "int64")),
        (ibis.literal(2, "int16"), ibis.literal(3, "int8")),
    ],
)
def test_extension_substring(compiler, two, three):
    t = ibis.table([("one", "string"), ("two", "float"), ("three", "int32")], name="t")

    query = t[t.one == t.one.substr(two, three)]
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    uris = [ext.uri for ext in plan.extension_uris]

    assert "equal:any_any" in scalar_func_names
    assert "substring:str_i32_i32" in scalar_func_names

    assert f"{DEFAULT_PREFIX}/functions_comparison.yaml" in uris
    assert f"{DEFAULT_PREFIX}/functions_string.yaml" in uris


@pytest.mark.parametrize(
    "left, right, bin_op, exp_func",
    [
        ("int32", "int32", operator.gt, "gt:any_any"),
        ("int32", "int64", operator.lt, "lt:any_any"),
        ("int16", "int8", operator.ge, "gte:any_any"),
        ("float64", "int8", operator.le, "lte:any_any"),
        ("float64", "float32", operator.eq, "equal:any_any"),
        ("int8", "float64", operator.ne, "not_equal:any_any"),
    ],
)
def test_extension_binop_comparison(compiler, left, right, bin_op, exp_func):
    t = ibis.table([("left", left), ("right", right)], name="t")

    query = t[bin_op(t.left, t.right)]
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    uris = [ext.uri for ext in plan.extension_uris]

    assert exp_func in scalar_func_names

    assert f"{DEFAULT_PREFIX}/functions_comparison.yaml" in uris


@pytest.mark.parametrize(
    "left, unary_op, exp_func",
    [
        ("int32", ops.NotNull, "is_not_null"),
        ("string", ops.IsNull, "is_null"),
        ("float64", ops.IsInf, "is_infinite"),
        ("float32", ops.IsNan, "is_nan"),
    ],
)
def test_extension_unaryop_comparison(compiler, left, unary_op, exp_func):
    t = ibis.table([("left", left)], name="t")

    query = t[unary_op(t.left).to_expr()]
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    uris = [ext.uri for ext in plan.extension_uris]

    assert any(exp_func in func_name for func_name in scalar_func_names)

    assert f"{DEFAULT_PREFIX}/functions_comparison.yaml" in uris


@pytest.mark.parametrize(
    "left, right, bin_op, exp_func, exp_uri",
    [
        (
            "decimal(15, 3)",
            "decimal(15, 3)",
            operator.add,
            "add:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "decimal(15, 3)",
            "decimal(15, 3)",
            operator.sub,
            "subtract:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "decimal(15, 3)",
            "decimal(15, 3)",
            operator.mul,
            "multiply:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "decimal(15, 3)",
            "decimal(15, 3)",
            operator.truediv,
            "divide:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "decimal(15, 3)",
            "int",
            operator.add,
            # should promote to dec_dec from dec_int
            "add:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "float64",
            "decimal(15, 3)",
            operator.sub,
            # should promote to dec_dec from fp32_dec
            "subtract:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "decimal(15, 3)",
            "int32",
            operator.mul,
            # should promote to dec_dec from dec_int
            "multiply:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "float32",
            "decimal(15, 3)",
            operator.truediv,
            # should promote to dec_dec from fp32_dec
            "divide:dec_dec",
            f"{DEFAULT_PREFIX}/functions_arithmetic_decimal.yaml",
        ),
        (
            "int",
            "int8",
            operator.add,
            # int is shorthand for i64, should promote to i64_i64
            "add:i64_i64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "int32",
            "int",
            operator.sub,
            # int is shorthand for i64, should promote to i64_i64
            "subtract:i64_i64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "int64",
            "int",
            operator.mul,
            # int is shorthand for i64, should promote to i64_i64
            "multiply:i64_i64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "int",
            "int",
            operator.truediv,
            "divide:i64_i64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "float64",
            "float",
            operator.add,
            "add:fp64_fp64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "float",
            "float32",
            operator.sub,
            # float is shorthand for fp64, should promote to fp64_fp64
            "subtract:fp64_fp64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "float32",
            "float32",
            operator.sub,
            "subtract:fp32_fp32",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "float64",
            "float64",
            operator.mul,
            "multiply:fp64_fp64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
        (
            "float",
            "float",
            operator.truediv,
            "divide:fp64_fp64",
            f"{DEFAULT_PREFIX}/functions_arithmetic.yaml",
        ),
    ],
)
def test_extension_arithmetic(compiler, left, right, bin_op, exp_func, exp_uri):
    t = ibis.table([("left", left), ("right", right)], name="t")

    query = t.mutate(new=bin_op(t.left, t.right))
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    uris = [ext.uri for ext in plan.extension_uris]

    assert exp_func in scalar_func_names

    assert exp_uri in uris


@pytest.mark.parametrize(
    "left, right, bin_op, exp_func, exp_uri",
    [
        (
            "bool",
            "bool",
            operator.and_,
            "and:bool",  # `and` is variadic with minimum 0 args so
            # expected signature is a single `bool`
            f"{DEFAULT_PREFIX}/functions_boolean.yaml",
        ),
        (
            "bool",
            "bool",
            operator.or_,
            "or:bool",  # `or` is variadic with minimum 0 args so
            # expected signature is a single `bool`
            f"{DEFAULT_PREFIX}/functions_boolean.yaml",
        ),
    ],
)
def test_extension_boolean(compiler, left, right, bin_op, exp_func, exp_uri):
    t = ibis.table([("left", left), ("right", right)], name="t")

    query = t.mutate(new=bin_op(t.left, t.right))
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    uris = [ext.uri for ext in plan.extension_uris]

    assert exp_func in scalar_func_names

    assert exp_uri in uris


def test_extension_udf_compile(compiler):
    try:
        from ibis.udf.vectorized import elementwise
    except ImportError:
        from ibis.legacy.udf.vectorized import elementwise

    pc = None

    t = ibis.table([("a", "int")], name="t")

    @elementwise(input_type=["int64"], output_type="int64")
    def add1(col, ctx=None):
        return pc.call_function("add", [col, 1], memory_pool=ctx.memory_pool)

    @elementwise(input_type=["int64"], output_type="int64")
    def sub1(col, ctx=None):
        return pc.call_function("subtract", [col, 1], memory_pool=ctx.memory_pool)

    query = t.mutate(b=add1(t.a), c=sub1(t.a))

    with pytest.raises(ValueError, match="udf_uri"):
        plan = compiler.compile(query)

    compiler.udf_uri = "orkbork"

    plan = compiler.compile(query)

    assert plan.extension_uris[0].uri == "orkbork"
    assert len(plan.extensions) == 2
    assert plan.extensions[0].extension_function.name == "add1"
    assert plan.extensions[1].extension_function.name == "sub1"


def test_extension_register_uri_override(tmp_path):
    from ibis_substrait.compiler.mapping import (
        _extension_mapping,
        register_extension_yaml,
    )

    sample_yaml = """scalar_functions:
  -
    name: "anotheradd"
    impls:
      - args:
          - name: x
            value: a
          - name: y
            value: b
        return: c"""

    yaml_file = tmp_path / "foo.yaml"
    yaml_file.write_text(sample_yaml)

    register_extension_yaml(yaml_file, uri="orkbork")

    assert _extension_mapping["anotheradd"]
    assert _extension_mapping["anotheradd"][(("a", "b"), "c")].uri == "orkbork"

    register_extension_yaml(yaml_file, prefix="orkbork")
    assert _extension_mapping["anotheradd"]
    assert _extension_mapping["anotheradd"][(("a", "b"), "c")].uri == "orkbork/foo.yaml"


def test_extension_arithmetic_multiple_signatures(compiler):
    t = ibis.table([("left", "int64"), ("right", "float32")], name="t")

    query = t.mutate(
        intadd=t.left + t.left,
        floatadd=t.right + t.right,
        intsub=t.left - t.left,
        floatsub=t.right - t.right,
    )
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    assert "add:i64_i64" in scalar_func_names
    assert "add:fp32_fp32" in scalar_func_names
    assert "subtract:i64_i64" in scalar_func_names
    assert "subtract:fp32_fp32" in scalar_func_names


_TYPE_MAPPING = {
    "int8": "i8",
    "int16": "i16",
    "int32": "i32",
    "int64": "i64",
    "float32": "fp32",
    "float64": "fp64",
}


@pytest.mark.parametrize(
    "col_dtype", ["float32", "float64", "int8", "int16", "int32", "int64"]
)
@pytest.mark.parametrize("digits_dtype", ["int8", "int16", "int32", "int64"])
def test_extension_round_upcast(compiler, col_dtype, digits_dtype):
    t = ibis.table([("col", col_dtype)], name="t")

    query = t.mutate(col=t.col.round(ibis.literal(8, type=digits_dtype)))
    plan = compiler.compile(query)

    scalar_func_names = [
        extension.extension_function.name for extension in plan.extensions
    ]

    assert f"round:{_TYPE_MAPPING[col_dtype]}_i32" in scalar_func_names


def test_ops_mapping_validity():
    from ibis_substrait.compiler import translate
    from ibis_substrait.compiler.mapping import (
        IBIS_SUBSTRAIT_OP_MAPPING,
        _extension_mapping,
    )

    for op in IBIS_SUBSTRAIT_OP_MAPPING.keys():
        assert hasattr(ops, op) or hasattr(translate, op)

    # `any` isn't a valid mapping
    for target in IBIS_SUBSTRAIT_OP_MAPPING.values():
        if target == "any":
            # any is special-cased
            continue
        assert target in _extension_mapping.keys()



from abc import ABC, abstractmethod

import datafusion
import pyarrow as pa


class SubstraitConsumer(ABC):
    @abstractmethod
    def with_tables(self, datasets: dict[str, pa.Table]):
        pass

    @abstractmethod
    def execute(self, plan) -> pa.Table:
        pass


class AceroSubstraitConsumer(SubstraitConsumer):
    def __init__(self) -> None:
        super().__init__()

    def with_tables(self, datasets: dict[str, pa.Table]):
        self.datasets = datasets
        return self

    def execute(self, plan) -> pa.Table:
        import pyarrow.substrait as pa_substrait

        def get_table_provider(datasets):
            def table_provider(names, schema):
                return datasets[names[0]]

            return table_provider

        query_bytes = plan.SerializeToString()
        result = pa_substrait.run_query(
            pa.py_buffer(query_bytes),
            table_provider=get_table_provider(self.datasets),
        )

        return result.read_all()


class DatafusionSubstraitConsumer(SubstraitConsumer):
    def __init__(self) -> None:
        self.connection = datafusion.SessionContext()

    def with_tables(self, datasets: dict[str, pa.Table]):
        for k, v in datasets.items():
            self.connection.deregister_table(k)
            self.connection.register_record_batches(k, [v.to_batches()])
        return self

    def execute(self, plan) -> pa.Table:
        plan_data = plan.SerializeToString()
        substrait_plan = datafusion.substrait.serde.deserialize_bytes(plan_data)
        logical_plan = datafusion.substrait.consumer.from_substrait_plan(
            self.connection, substrait_plan
        )

        df = self.connection.create_dataframe_from_logical_plan(logical_plan)
        for column_number, column_name in enumerate(df.schema().names):
            df = df.with_column_renamed(
                column_name, plan.relations[0].root.names[column_number]
            )
        record_batch = df.collect()
        return pa.Table.from_batches(record_batch)



import os
import tempfile

import duckdb
import ibis
import pyarrow as pa
import pyarrow.compute as pc
import pytest
from ibis import _
from ibis.conftest import LINUX, SANDBOXED

from ibis_substrait.compiler.core import SubstraitCompiler

from .parity_utils import (
    AceroSubstraitConsumer,
    DatafusionSubstraitConsumer,
    SubstraitConsumer,
)

nix_linux_sandbox = pytest.mark.xfail(
    LINUX and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
    raises=duckdb.IOException,
)


def sort_pyarrow_table(table: pa.Table):
    sort_keys = [(name, "ascending") for name in table.column_names]
    sort_indices = pc.sort_indices(table, sort_keys)
    return pc.take(table, sort_indices)


# TODO move this into a consumer class
def run_query_duckdb(query, datasets):
    with tempfile.TemporaryDirectory() as tempdir:
        con = ibis.duckdb.connect(os.path.join(tempdir, "temp.db"))
        for table_name, pa_table in datasets.items():
            con.create_table(name=table_name, obj=ibis.memtable(pa_table))

        # TODO con.to_pyarrow(query) in duckdb backend doesn't work with latest ibis and pyarrow versions
        res = pa.Table.from_pandas(con.to_pandas(query))
        con.disconnect()
        return res


orders_raw = [
    ("order_id", "int64", [1, 2, 3, 4]),
    ("fk_store_id", "int64", [1, 1, 2, 2]),
    ("fk_customer_id", "int64", [10, 11, 13, 13]),
    ("description", "string", ["A", "B", "C", "D"]),
    ("order_total", "float", [10.0, 32.3, 32.0, 140.0]),
]

stores_raw = [("store_id", "int64", [1, 2, 3]), ("city", "string", ["NY", "LA", "NY"])]

customers_raw = [
    ("customer_id", "int64", [10, 11, 13]),
    ("name", "string", ["Ann", "Bob", "Chris"]),
]

orders = ibis.table([(x[0], x[1]) for x in orders_raw], name="orders")
stores = ibis.table([(x[0], x[1]) for x in stores_raw], name="stores")
customers = ibis.table([(x[0], x[1]) for x in customers_raw], name="customers")

datasets = {
    "orders": pa.Table.from_pydict({x[0]: x[2] for x in orders_raw}),
    "stores": pa.Table.from_pydict({x[0]: x[2] for x in stores_raw}),
    "customers": pa.Table.from_pydict({x[0]: x[2] for x in customers_raw}),
}


@pytest.fixture(scope="session")
def acero_consumer():
    return AceroSubstraitConsumer().with_tables(datasets)


@pytest.fixture(scope="session")
def datafusion_consumer():
    return DatafusionSubstraitConsumer().with_tables(datasets)


def run_parity_test(consumer: SubstraitConsumer, expr):
    res_duckdb = sort_pyarrow_table(run_query_duckdb(expr, datasets))

    compiler = SubstraitCompiler()

    res_compare = sort_pyarrow_table(consumer.execute(compiler.compile(expr)))

    assert res_compare.equals(res_duckdb)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_projection(consumer: str, request):
    expr = orders["order_id", "order_total"]
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_mutate(consumer: str, request):
    expr = orders.mutate(order_total_plus_1=orders["order_total"] + 1)
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_sort(consumer: str, request):
    expr = orders.order_by("order_total")
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_sort_limit(consumer: str, request):
    expr = orders.order_by("order_total").limit(2)
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_filter(consumer: str, request):
    expr = orders.filter(lambda t: t.order_total > 30)
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_inner_join(consumer: str, request):
    expr = orders.join(stores, orders["fk_store_id"] == stores["store_id"])
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_inner_join_projection(consumer: str, request):
    expr = orders.join(stores, orders["fk_store_id"] == stores["store_id"]).select(
        "store_id"
    )
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        pytest.param(
            "acero_consumer",
            marks=[
                pytest.mark.xfail(pa.ArrowNotImplementedError, reason="Unimplemented")
            ],
        ),
        "datafusion_consumer",
    ],
)
def test_cross_join(consumer: str, request):
    expr = orders.cross_join(stores)
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_left_join(consumer: str, request):
    expr = orders.join(stores, orders["fk_store_id"] == stores["store_id"], how="left")
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    ["acero_consumer", "datafusion_consumer"],
)
def test_filter_groupby(consumer: str, request):
    filter_table = orders.join(
        stores, orders["fk_store_id"] == stores["store_id"]
    ).filter(lambda t: t.order_total > 30)

    expr = (
        filter_table.group_by("city")
        .aggregate(sales=filter_table["order_id"].count())
        .filter(ibis.literal(True))
    )

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        pytest.param(
            "datafusion_consumer",
            marks=[pytest.mark.xfail(Exception, reason="")],
        ),
    ],
)
def test_groupby_datafusion(consumer: str, request):
    expr = orders.group_by("fk_store_id").aggregate(sales=orders["order_id"].count())

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        pytest.param(
            "acero_consumer",
            marks=[
                pytest.mark.xfail(pa.ArrowNotImplementedError, reason="Unimplemented")
            ],
        ),
        "datafusion_consumer",
    ],
)
def test_filter_groupby_count_distinct(consumer: str, request):
    filter_table = orders.join(
        stores, orders["fk_store_id"] == stores["store_id"]
    ).filter(lambda t: t.order_total > 30)

    expr = (
        filter_table.group_by("city")
        .aggregate(sales=filter_table["city"].nunique())
        .filter(ibis.literal(True))
    )

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    ["acero_consumer", "datafusion_consumer"],
)
def test_aggregate_having(consumer: str, request):
    expr = orders.aggregate(
        [orders.order_id.max().name("amax"), orders.order_id.count().name("acount")],
        by="fk_store_id",
        having=(_.order_id.count() > 1),
    )

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_inner_join_chain(consumer: str, request):
    expr = orders.join(stores, orders["fk_store_id"] == stores["store_id"]).join(
        customers, orders["fk_customer_id"] == customers["customer_id"]
    )

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_union(consumer: str, request):
    expr = orders.union(orders)

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        pytest.param(
            "acero_consumer",
            marks=[
                pytest.mark.xfail(pa.ArrowNotImplementedError, reason="Unimplemented")
            ],
        ),
        "datafusion_consumer",
    ],
)
def test_window(consumer: str, request):
    expr = orders.select(
        orders["order_total"].mean().over(ibis.window(group_by="fk_store_id"))
    )

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_is_in(consumer: str, request):
    expr = stores.filter(stores.city.isin(["NY", "LA"]))

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        pytest.param(
            "acero_consumer",
            marks=[
                pytest.mark.xfail(pa.ArrowNotImplementedError, reason="Unimplemented")
            ],
        ),
        "datafusion_consumer",
    ],
)
def test_scalar_subquery(consumer: str, request):
    expr = orders.filter(orders["order_total"] == orders["order_total"].max())

    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize("consumer", ["acero_consumer", "datafusion_consumer"])
def test_projection_functions_arithmetic(consumer: str, request):
    expr = orders.select(
        orders["order_id"] + orders["order_total"],
        orders["order_id"] - orders["order_total"],
        orders["order_id"] * orders["order_total"],
        orders["order_id"] / orders["order_total"],
        orders["order_total"] ** ibis.literal(2),
        orders["order_total"].sqrt(),
        orders["order_total"].exp(),
        (orders["order_id"] * 10 - orders["order_total"]).abs(),
    )
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        "acero_consumer",
        pytest.param(
            "datafusion_consumer",
            marks=[pytest.mark.xfail(Exception, reason="NotImplemented")],
        ),
    ],
)
def test_projection_functions_arithmetic_negation(consumer: str, request):
    expr = orders.select(orders["order_total"].negate())
    run_parity_test(request.getfixturevalue(consumer), expr)


@pytest.mark.parametrize(
    "consumer",
    [
        pytest.param(
            "acero_consumer",
            marks=[
                pytest.mark.xfail(
                    pa.ArrowNotImplementedError,
                    reason="No conversion function exists to convert the Substrait function",
                )
            ],
        ),
        pytest.param(
            "datafusion_consumer",
            marks=[pytest.mark.xfail(Exception, reason="NotImplemented")],
        ),
    ],
)
def test_projection_functions_arithmetic_modulus(consumer: str, request):
    expr = orders.select(
        orders["order_id"] % orders["fk_store_id"],
    )
    run_parity_test(request.getfixturevalue(consumer), expr)



from collections import OrderedDict

import ibis
import ibis.expr.datatypes as dt
import pytest
from google.protobuf import json_format
from ibis import _
from substrait.gen.proto import algebra_pb2 as stalg
from substrait.gen.proto import type_pb2 as stt

from ibis_substrait.compiler.translate import translate

NULLABILITY_NULLABLE = stt.Type.Nullability.NULLABILITY_NULLABLE
NULLABILITY_REQUIRED = stt.Type.Nullability.NULLABILITY_REQUIRED


@pytest.fixture
def t():
    return ibis.table(
        [
            ("full_name", "string"),
            ("age", "int64"),
            ("ts", "timestamp('UTC')"),
            ("delta", "interval"),
        ]
    )


def to_dict(message):
    return json_format.MessageToDict(message)


def test_aggregation(t, compiler):
    expr = (
        t.group_by(name_len=lambda t: t.full_name.length())
        .aggregate(max_age=t.age.max(), min_age=t.age.min())
        .filter(lambda t: t.name_len > 3)
    )
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


def test_aggregation_with_sort(t, compiler):
    expr = (
        t.group_by(name_len=lambda t: t.full_name.length())
        .aggregate(max_age=t.age.max(), min_age=t.age.min())
        .order_by(_.max_age)
        .filter(lambda t: t.name_len > 3)
    )
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


def test_aggregation_no_args(t, compiler):
    expr = t.group_by("age").count()
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


@pytest.mark.parametrize(
    "preceding, following",
    [
        ((4, 2), None),
        (None, (1, 5)),
        (None, None),
        (2, 4),
    ],
)
def test_aggregation_window(t, compiler, preceding, following):
    expr = t.projection(
        [
            t.full_name.length()
            .mean()
            .over(ibis.window(preceding=preceding, following=following, group_by="age"))
        ]
    )
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


def test_array_literal(compiler):
    expr = ibis.literal(["a", "b"])
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


def test_map_literal(compiler):
    expr = ibis.literal(dict(a=[], b=[2]))
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


def test_struct_literal(compiler):
    expr = ibis.literal(OrderedDict(a=1, b=[2.0]))
    result = translate(expr, compiler=compiler)
    js = to_dict(result)
    assert js


def test_translate_table_expansion(compiler):
    t = ibis.table([("a", "int32"), ("b", "int64")], name="table0")
    expr = t.mutate(c=t.a + t.b)
    result = translate(expr, compiler=compiler)
    expected = {
        "project": {
            "common": {
                "emit": {"outputMapping": [2, 3, 4]},
                "hint": {"outputNames": ["a", "b", "c"]},
            },
            "input": {
                "read": {
                    "common": {"direct": {}, "hint": {"outputNames": ["a", "b"]}},
                    "baseSchema": {
                        "names": ["a", "b"],
                        "struct": {
                            "nullability": "NULLABILITY_REQUIRED",
                            "types": [
                                {"i32": {"nullability": "NULLABILITY_NULLABLE"}},
                                {"i64": {"nullability": "NULLABILITY_NULLABLE"}},
                            ],
                        },
                    },
                    "namedTable": {"names": ["table0"]},
                }
            },
            "expressions": [
                {
                    "selection": {
                        "directReference": {"structField": {}},
                        "rootReference": {},
                    }
                },
                {
                    "selection": {
                        "directReference": {"structField": {"field": 1}},
                        "rootReference": {},
                    }
                },
                {
                    "scalarFunction": {
                        "functionReference": 1,
                        "arguments": [
                            {
                                "value": {
                                    "cast": {
                                        "failureBehavior": "FAILURE_BEHAVIOR_THROW_EXCEPTION",
                                        "input": {
                                            "selection": {
                                                "directReference": {"structField": {}},
                                                "rootReference": {},
                                            },
                                        },
                                        "type": {
                                            "i64": {
                                                "nullability": "NULLABILITY_NULLABLE"
                                            }
                                        },
                                    },
                                },
                            },
                            {
                                "value": {
                                    "selection": {
                                        "directReference": {
                                            "structField": {"field": 1},
                                        },
                                        "rootReference": {},
                                    },
                                },
                            },
                        ],
                        "outputType": {"i64": {"nullability": "NULLABILITY_NULLABLE"}},
                    }
                },
            ],
        }
    }
    assert to_dict(result) == expected


def test_emit_mutate_select_all(compiler):
    t = ibis.table([("a", "int64"), ("b", "char"), ("c", "int32")], name="table0")
    expr = t.mutate(d=t.a + 1)
    result = translate(expr, compiler=compiler)
    expected = {
        "project": {
            "common": {
                "emit": {"outputMapping": [3, 4, 5, 6]},
                "hint": {"outputNames": ["a", "b", "c", "d"]},
            },
            "input": {
                "read": {
                    "common": {"direct": {}, "hint": {"outputNames": ["a", "b", "c"]}},
                    "baseSchema": {
                        "names": ["a", "b", "c"],
                        "struct": {
                            "types": [
                                {"i64": {"nullability": "NULLABILITY_NULLABLE"}},
                                {"string": {"nullability": "NULLABILITY_NULLABLE"}},
                                {"i32": {"nullability": "NULLABILITY_NULLABLE"}},
                            ],
                            "nullability": "NULLABILITY_REQUIRED",
                        },
                    },
                    "namedTable": {"names": ["table0"]},
                }
            },
            "expressions": [
                {
                    "selection": {
                        "directReference": {"structField": {}},
                        "rootReference": {},
                    }
                },
                {
                    "selection": {
                        "directReference": {"structField": {"field": 1}},
                        "rootReference": {},
                    }
                },
                {
                    "selection": {
                        "directReference": {"structField": {"field": 2}},
                        "rootReference": {},
                    }
                },
                {
                    "scalarFunction": {
                        "functionReference": 1,
                        "outputType": {"i64": {"nullability": "NULLABILITY_NULLABLE"}},
                        "arguments": [
                            {
                                "value": {
                                    "selection": {
                                        "directReference": {"structField": {}},
                                        "rootReference": {},
                                    }
                                },
                            },
                            {
                                "value": {
                                    "cast": {
                                        "failureBehavior": "FAILURE_BEHAVIOR_THROW_EXCEPTION",
                                        "input": {"literal": {"i8": 1}},
                                        "type": {
                                            "i64": {
                                                "nullability": "NULLABILITY_NULLABLE"
                                            }
                                        },
                                    }
                                }
                            },
                        ],
                    }
                },
            ],
        }
    }

    assert to_dict(result) == expected


def test_emit_nested_projection_output_mapping(compiler):
    t = ibis.table(
        [
            ("a", "int64"),
            ("b", "int64"),
            ("c", "int64"),
            ("d", "int64"),
        ],
        name="t",
    )
    expr = t["a", "b", "c", "d"]
    result = translate(expr, compiler=compiler)
    # root table has 4 columns, so output mapping starts at index 4
    # should have 4 entries
    assert result.project.common.emit.output_mapping == [4, 5, 6, 7]

    expr = expr["a", "b", "c"]
    result = translate(expr, compiler=compiler)
    # previous emit has 4 columns, so output mapping starts at index 4
    # should have 3 entries
    assert result.project.common.emit.output_mapping == [4, 5, 6]

    expr = expr["a", "b"]
    result = translate(expr, compiler=compiler)
    # previous emit has 3 columns, so output mapping starts at index 3
    # should have 2 entries
    assert result.project.common.emit.output_mapping == [3, 4]


def test_aggregate_project_output_mapping(compiler):
    t = ibis.table(
        [
            ("a", "int64"),
            ("b", "int64"),
            ("c", "int64"),
            ("d", "int64"),
            ("e", "int64"),
        ],
        name="t",
    )

    expr = (
        t.group_by(["a", "b"]).aggregate([t.c.sum().name("sum")]).select(["b", "sum"])
    )

    result = translate(expr, compiler=compiler)
    assert result.project.common.emit.output_mapping == [3, 4]

    # Select 3 of 5 columns in base table to make sure output mapping reflects
    # a count from the outputs of the aggregation ('a', 'b', and 'sum')
    expr = (
        t.select(["a", "b", "c"])
        .group_by(["a", "b"])
        .aggregate([t.c.sum().name("sum")])
        .select(["b", "sum"])
    )

    result = translate(expr, compiler=compiler)
    assert result.project.common.emit.output_mapping == [3, 4]


def test_ibis_schema_to_substrait_schema():
    input = ibis.schema(
        [
            (
                "a",
                dt.Array(dt.Struct.from_tuples([("b", "int64"), ("c", "int64")])),
            ),
            (
                "b",
                dt.Array(dt.Struct.from_tuples([("b", "int64"), ("c", "int64")])),
            ),
            ("d", "int64"),
        ]
    )

    expected = stt.NamedStruct(
        names=["a", "b", "c", "b", "b", "c", "d"],
        struct=stt.Type.Struct(
            types=[
                stt.Type(
                    list=stt.Type.List(
                        type=stt.Type(
                            struct=stt.Type.Struct(
                                types=[
                                    stt.Type(
                                        i64=stt.Type.I64(
                                            nullability=NULLABILITY_NULLABLE,
                                        )
                                    ),
                                    stt.Type(
                                        i64=stt.Type.I64(
                                            nullability=NULLABILITY_NULLABLE,
                                        )
                                    ),
                                ],
                                nullability=NULLABILITY_NULLABLE,
                            )
                        ),
                        nullability=NULLABILITY_NULLABLE,
                    )
                ),
                stt.Type(
                    list=stt.Type.List(
                        type=stt.Type(
                            struct=stt.Type.Struct(
                                types=[
                                    stt.Type(
                                        i64=stt.Type.I64(
                                            nullability=NULLABILITY_NULLABLE,
                                        )
                                    ),
                                    stt.Type(
                                        i64=stt.Type.I64(
                                            nullability=NULLABILITY_NULLABLE,
                                        )
                                    ),
                                ],
                                nullability=NULLABILITY_NULLABLE,
                            )
                        ),
                        nullability=NULLABILITY_NULLABLE,
                    )
                ),
                stt.Type(i64=stt.Type.I64(nullability=NULLABILITY_NULLABLE)),
            ],
            nullability=NULLABILITY_REQUIRED,
        ),
    )
    assert translate(input) == expected


@pytest.mark.parametrize(("name", "expected_offset"), [("a", 0), ("b", 1), ("c", 2)])
def test_simple_field_access(compiler, name, expected_offset):
    t = ibis.table(
        dict(
            a="string",
            b="struct<a: string, b: int64, c: float64>",
            c="array<array<float64>>",
        )
    )
    expr = t[name]
    expected = json_format.ParseDict(
        {
            "selection": {
                "direct_reference": {"struct_field": {"field": expected_offset}},
                "root_reference": {},
            }
        },
        stalg.Expression(),
    )
    result = translate(expr, compiler=compiler)
    assert result == expected


@pytest.mark.parametrize(("name", "expected_offset"), [("a", 0), ("b", 1), ("c", 2)])
def test_struct_field_access(compiler, name, expected_offset):
    t = ibis.table(dict(f="struct<a: string, b: int64, c: float64>"))
    expr = t.f[name]
    expected = json_format.ParseDict(
        {
            "selection": {
                "direct_reference": {
                    "struct_field": {
                        "field": 0,
                        "child": {"struct_field": {"field": expected_offset}},
                    }
                },
                "root_reference": {},
            }
        },
        stalg.Expression(),
    )
    result = translate(expr, compiler=compiler)
    assert result == expected


def test_nested_struct_field_access(compiler):
    t = ibis.table(
        dict(f="struct<a: struct<a: int64, b: struct<a: int64, b: int64, c: int64>>>")
    )
    #        0  0    1    2
    expr = t.f["a"]["b"]["c"]
    expected = json_format.ParseDict(
        {
            "selection": {
                "direct_reference": {
                    "struct_field": {
                        "field": 0,
                        "child": {
                            "struct_field": {
                                "field": 0,
                                "child": {
                                    "struct_field": {
                                        "field": 1,
                                        "child": {"struct_field": {"field": 2}},
                                    }
                                },
                            }
                        },
                    }
                },
                "root_reference": {},
            }
        },
        stalg.Expression(),
    )
    result = translate(expr, compiler=compiler)
    assert result == expected


def test_function_argument_usage(compiler):
    t = ibis.table([("a", "int64")], name="t")
    expr = t.a.count()

    result = translate(expr, compiler=compiler)
    # Check that there is an `arguments` field
    # and that it has the expected `value`
    expected = json_format.ParseDict(
        {
            "value": {
                "selection": {
                    "direct_reference": {"struct_field": {}},
                    "root_reference": {},
                },
            },
        },
        stalg.FunctionArgument(),
    )

    assert result.arguments[0] == expected


def test_aggregate_filter_select_output_mapping(compiler):
    t = ibis.table([("a", "int"), ("b", "float"), ("c", "int")], name="t")

    expr = (
        t.aggregate(
            [
                t.a.max().name("amax"),
                t.b.max().name("bmax"),
                t.a.min().name("amin"),
                t.c.min().name("cmin"),
                t.c.max().name("cmax"),
            ],
        )
        .filter(_.amin < 5)
        .select("amin", "bmax")
    )

    # Aggregate of 3 columns with 5 created columns as the output
    # Filter to put an op between aggregate and select
    # Project out two columns, output_mapping should be 5, 6
    # because 0, 1, 2, 3, 4 correspond to the created aggregate cols.
    result = translate(expr, compiler=compiler)

    assert result.project.common.emit.output_mapping == [5, 6]


def test_filter_over_subquery(compiler):
    t = ibis.table([("a", "int")], name="t").filter(_.a > _.a.mean())
    translate(t, compiler=compiler)


def test_groupby_multiple_keys(compiler):
    t = ibis.table(name="t", schema=(("a", "int"), ("b", "int")))
    expr = t.group_by(["a", "b"]).agg()
    plan = translate(expr, compiler=compiler)

    # There should be one grouping with two separate expressions inside
    assert len(plan.aggregate.groupings) == 1
    assert len(plan.aggregate.groupings[0].grouping_expressions) == 2


def test_join_chain_indexing_in_group_by(compiler):
    t1 = ibis.table([("a", int), ("b", int)], name="t1")
    t2 = ibis.table([("a", int), ("c", int)], name="t2")
    t3 = ibis.table([("a", int), ("d", int)], name="t3")
    t4 = ibis.table([("a", int), ("c", int)], name="t4")

    join_chain = t1.join(t2, "a").join(t3, "a").join(t4, "a")
    # Indexing for chained join
    # t1: a: 0
    # t1: b: 1
    # t2: a: 2
    # t2: c: 3
    # t3: a: 4
    # t3: d: 5
    # t4: a: 6
    # t4: c: 7

    expr = join_chain.group_by("d").count().select("d")
    plan = compiler.compile(expr)
    # Check that the field index for the group_by key is correctly indexed
    assert (
        plan.relations[0]
        .root.input.project.input.aggregate.groupings[0]
        .grouping_expressions[0]
        .selection.direct_reference.struct_field.field
        == 3
    )

    expr = join_chain.group_by("c").count().select("c")
    plan = compiler.compile(expr)
    # Check that the field index for the group_by key is correctly indexed
    assert (
        plan.relations[0]
        .root.input.project.input.aggregate.groupings[0]
        .grouping_expressions[0]
        .selection.direct_reference.struct_field.field
        == 2
    )

    # Group-by on a column that will be renamed by the joinchain
    expr = join_chain.group_by(t4.c).count().select("c")
    plan = compiler.compile(expr)
    # Check that the field index for the group_by key is correctly indexed
    assert (
        plan.relations[0]
        .root.input.project.input.aggregate.groupings[0]
        .grouping_expressions[0]
        .selection.direct_reference.struct_field.field
        == 4
    )


_window_hows = {
    "unspecified": "BOUNDS_TYPE_UNSPECIFIED",
    "range": "BOUNDS_TYPE_RANGE",
    "rows": "BOUNDS_TYPE_ROWS",
}


@pytest.mark.parametrize(
    "bounds",
    [
        (-4, 2),
        (1, 5),
        (None, None),
        (2, 4),
    ],
)
@pytest.mark.parametrize("how", ["range", "rows"])
def test_aggregation_window_how(t, compiler, bounds, how):
    how_arg = {how: bounds}
    expr = t.projection(
        [t.full_name.length().mean().over(ibis.window(group_by="age", **how_arg))]
    )
    result = translate(expr, compiler=compiler)

    bounds_type_int = result.project.expressions[0].window_function.bounds_type

    assert (
        stalg.Expression.WindowFunction.BoundsType.Name(bounds_type_int)
        == _window_hows[how]
    )
